import dataclasses
from enum import Enum
from functools import cached_property
from typing import Dict, Hashable, List, Optional, Tuple

import numpy as np
import shapely
import xarray as xr
from emsarray.conventions import Convention, Specificity
from emsarray.exceptions import ConventionViolationError
from emsarray.types import Pathish
from shapely.geometry.base import BaseGeometry


class SMCGridKind(str, Enum):
    cell = 'cell'


SMCIndex = Tuple[SMCGridKind, int]


@dataclasses.dataclass
class SMCTopology:
    dataset: xr.Dataset

    #: The name of the global attribute that names the cell dimension
    cell_dimension_attribute = 'SMC_grid_type'

    def __init__(
        self,
        dataset: xr.Dataset,
        *,
        cell_dimension: Optional[Hashable] = None,
        longitude: Optional[Hashable] = None,
        latitude: Optional[Hashable] = None,
        longitude_cell_size_factor: Optional[Hashable] = None,
        latitude_cell_size_factor: Optional[Hashable] = None,
    ):
        """
        Construct a new :class:`CFGridTopology` instance for a dataset.

        By default this will introspect the dataset
        looking for longitude and latitude coordinate variables,
        and lookinf for longitude and latitude cell size variables.
        The ``longitude``, ``latitude``,
        ``longitude_cell_size``, and ``latitude_cell_size`` parameters
        allow you to manually specify the correct variable names
        if the automatic detection fails.
        """
        self.dataset = dataset
        if cell_dimension is not None:
            self.cell_dimension = cell_dimension
        if longitude is not None:
            self.longitude_name = longitude
        if latitude is not None:
            self.latitude_name = latitude
        if longitude_cell_size_factor is not None:
            self.longitude_cell_size_factor_name = longitude_cell_size_factor
        if latitude_cell_size_factor is not None:
            self.latitude_cell_size_factor_name = latitude_cell_size_factor

    @cached_property
    def cell_dimension(self) -> Hashable:
        """The dimension name that indexes each cell"""
        try:
            name = self.dataset.attrs[self.cell_dimension_attribute]
        except KeyError:
            raise ConventionViolationError(
                "Global cell dimension attribute "
                f"{self.cell_dimension_attribute!r} not set!")
        if name not in self.dataset.dims:
            raise ConventionViolationError(
                f"Cell dimension {name!r} does not exist in the dataset!")
        return name

    @property
    def cell_count(self) -> int:
        """The size of the cell dimension"""
        return self.dataset.dims[self.cell_dimension]

    @cached_property
    def longitude_name(self) -> Hashable:
        """
        The name of the longitude coordinate variable.
        Found by looking for a variable with either a
        ``standard_name = "longitude"`` or
        ``units = "degree_east"``
        attribute.
        """
        try:
            return next(
                name for name, variable in self.dataset.variables.items()
                if variable.attrs.get('standard_name') == 'longitude'
                or variable.attrs.get('coordinate_type') == 'longitude'
                or variable.attrs.get('units') == 'degree_east'
            )
        except StopIteration:
            raise ValueError("Could not find longitude coordinate")

    @cached_property
    def latitude_name(self) -> Hashable:
        """
        The name of the latitude coordinate variable.
        Found by looking for a variable with either a
        ``standard_name = "latitude"`` or
        ``units = "degree_north"``
        attribute.
        """
        try:
            return next(
                name
                for name, variable in self.dataset.variables.items()
                if variable.attrs.get('standard_name') == 'latitude'
                or variable.attrs.get('coordinate_type') == 'latitude'
                or variable.attrs.get('units') == 'degree_north'
            )
        except StopIteration:
            raise ValueError("Could not find latitude coordinate")

    @cached_property
    def longitude_cell_size_factor_name(self) -> Hashable:
        """
        The name of the longitude cell size variable.
        Found by looking for a variable with a
        ``long_name = "longitude cell size factor"`` attribute.
        """
        try:
            return next(
                name for name, variable in self.dataset.variables.items()
                if variable.attrs.get('long_name') == 'longitude cell size factor'
            )
        except StopIteration:
            raise ValueError("Could not find longitude cell size variable")

    @cached_property
    def latitude_cell_size_factor_name(self) -> Hashable:
        """
        The name of the latitude cell size variable.
        Found by looking for a variable with a
        ``long_name = "latitude cell size factor"`` attribute.
        """
        try:
            return next(
                name for name, variable in self.dataset.variables.items()
                if variable.attrs.get('long_name') == 'latitude cell size factor'
            )
        except StopIteration:
            raise ValueError("Could not find latitude cell size variable")

    @property
    def longitude(self) -> xr.DataArray:
        """The longitude coordinate variable"""
        return self.dataset[self.longitude_name]

    @property
    def latitude(self) -> xr.DataArray:
        """The latitude coordinate variable"""
        return self.dataset[self.latitude_name]

    @property
    def longitude_cell_size_factor(self) -> xr.DataArray:
        """The longitude cell size variable"""
        return self.dataset[self.longitude_cell_size_name]

    @property
    def latitude_cell_size_factor(self) -> xr.DataArray:
        """The latitude cell size variable"""
        return self.dataset[self.latitude_cell_size_name]


class SMC(Convention[SMCGridKind, int]):
    """
    Spherical multiple-cell (SMC) datasets consist of non overlapping,
    axis aligned, rectangular cells of varying sizes.
    Smaller cells are used where increased resolution is desired
    (e.g. around coastlines).
    Cells are indexed by the :attr:`SMC.cell_dimension` dimension,
    which is a one-dimensional index in an arbitrary order.
    """

    def __init__(
        self,
        dataset: xr.Dataset,
        *,
        topology: Optional[SMCTopology] = None,
    ):
        super().__init__(dataset)
        if topology is not None:
            self.topology = topology

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        # The following dataset attributes are required to identify this as an
        # SMC dataset
        required_attrs = [
            'base_lat_size', 'base_lon_size',
            'southernmost_latitude',
            'northernmost_latitude',
            'westernmost_longitude',
            'easternmost_longitude',
            SMCTopology.cell_dimension_attribute,
        ]

        if not all(attr in dataset.attrs for attr in required_attrs):
            return None

        return Specificity.HIGH

    @cached_property
    def topology(self) -> SMCTopology:
        return SMCTopology(self.dataset)

    def ravel_index(self, index: SMCIndex) -> int:
        _kind, linear_index = index
        return linear_index

    def unravel_index(
        self,
        linear_index: int,
        grid_kind: Optional[SMCGridKind] = None,
    ) -> SMCIndex:
        return (SMCGridKind.cell, linear_index)

    @property
    def grid_kinds(self) -> List[SMCGridKind]:
        return list(SMCGridKind)

    @property
    def default_grid_kind(self) -> SMCGridKind:
        return SMCGridKind.cell

    def get_grid_kind_and_size(
        self, data_array: xr.DataArray,
    ) -> Tuple[SMCGridKind, int]:
        if self.topology.cell_dimension not in data_array.dims:
            raise ValueError("Unknown grid kind")
        return (SMCGridKind.cell, self.topology.cell_count)

    def make_linear(self, data_array: xr.DataArray) -> xr.DataArray:
        kind, size = self.get_grid_kind_and_size(data_array)
        if kind is not SMCGridKind.cell:
            raise ValueError("Unknown grid kind")

        # The dataset is already linear
        return data_array

    def selector_for_index(self, index: SMCIndex) -> Dict[Hashable, int]:
        _kind, linear_index = index
        return {self.topology.cell_dimension: linear_index}

    def drop_geometry(self) -> xr.Dataset:
        # Drop geometry variables
        dataset = self.dataset.drop_vars([
            'longitude', 'latitude', 'cx', 'cy',
        ])

        # Drop geometry attributes
        required_attrs = [
            'base_lat_size', 'base_lon_size',
            'southernmost_latitude',
            'northernmost_latitude',
            'westernmost_longitude',
            'easternmost_longitude',
            'SMC_grid_type',
        ]
        for key in sorted(dataset.attrs.keys() & required_attrs):
            del dataset.attrs[key]

        return dataset

    @cached_property
    def polygons(self) -> np.ndarray:
        """
        SMC polygons are lat/lon boxes centred at a point, with a size given by
        cx/cy and the base cell size.
        """
        # In the SMC datasets I have seen, the step values have all been
        # exactly representable as floats - i.e. a power of two.
        # The following calculations are exact because of this.
        # If any SMC datasets are encountered that do _not_ use an exactly
        # representable power of two this will have to be modified.
        lon_size = float(self.dataset.attrs['base_lon_size'])
        lat_size = float(self.dataset.attrs['base_lat_size'])

        lons = self.dataset['longitude'].values
        lats = self.dataset['latitude'].values
        cx = self.dataset['cx'].values
        cy = self.dataset['cy'].values

        # Cells have size (cx * lon_size, cy * lat_size),
        # centred at (longitde, latitude)
        lon_cell_size = lon_size * cx / 2
        lat_cell_size = lat_size * cy / 2
        lon_min = lons - lon_cell_size
        lon_max = lons + lon_cell_size
        lat_min = lats - lat_cell_size
        lat_max = lats + lat_cell_size

        # points is an array of shape (cell_count, 5, 2),
        # where each row is a set of five points defining the cell polygon.
        points = np.array([
            [lon_min, lat_min],
            [lon_max, lat_min],
            [lon_max, lat_max],
            [lon_min, lat_max],
            [lon_min, lat_min],
        ], dtype=lons.dtype)
        points = np.transpose(points, (2, 0, 1))

        return shapely.polygons(points)

    def make_clip_mask(
        self,
        clip_geometry: BaseGeometry,
        buffer: int = 0,
    ) -> xr.Dataset:
        if buffer > 0:
            raise ValueError("Buffering SMC datasets is not yet implemented")

        spatial_index = self.spatial_index

        included_indices = [
            hit.linear_index
            for polygon, hit in spatial_index.query(clip_geometry)
            if polygon.intersects(clip_geometry)
            and not polygon.touches(clip_geometry)
        ]
        mask = np.zeros(self.topology.cell_count, dtype=bool)
        mask[included_indices] = True

        return xr.Dataset(
            data_vars={
                'mask': xr.DataArray(
                    data=mask,
                    dims=[self.topology.cell_dimension],
                ),
            },
        )

    def apply_clip_mask(self, clip_mask: xr.Dataset, work_dir: Pathish) -> xr.Dataset:
        return self.dataset.isel({self.topology.cell_dimension: clip_mask['mask'].values})



import pathlib

import numpy as np
import pytest
import pandas
import xarray
import xarray.testing
from emsarray.conventions import Specificity
from emsarray.conventions._registry import registry
from shapely.geometry import Polygon, box
import shapely.testing

from emsarray_smc import SMC, SMCGridKind

example_smc_file = pathlib.Path('./smc-australia.nc')


def test_convention_registered():
    assert SMC in registry.entry_point_conventions


def test_check_dataset_match(datasets):
    assert SMC.check_dataset(xarray.open_dataset(datasets / 'smc.nc')) is Specificity.HIGH


def test_open_dataset(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert isinstance(ds.ems, SMC)


def test_check_dataset_no_match(
    datasets: pathlib.Path,
):
    # Make a basic 1D CF grid dataset
    ds = xarray.Dataset(
        data_vars={'value': (['y', 'x'], np.arange(100).reshape(10, 10))},
        coords={
            'x': ('x', np.arange(10), {'standard_name': 'longitude'}),
            'y': ('y', np.arange(10), {'standard_name': 'latitude'}),
        },
    )

    assert SMC.check_dataset(ds) is None


def test_cell_dimension_autodetect(
    datasets: pathlib.Path,
):
    """
    The name of the cell dimension should be taken from the `SMC_grid_type`
    global attribute.
    """
    ds = xarray.Dataset(
        data_vars={'values': ('smc_dimension', np.arange(10))},
        attrs={'SMC_grid_type': 'smc_dimension'},
    )
    convention = SMC(ds)
    assert convention.topology.cell_dimension == 'smc_dimension'
    assert convention.topology.cell_count == 10


def test_ravel_index(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert ds.ems.ravel_index((SMCGridKind.cell, 10)) == 10


def test_unravel_index(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert ds.ems.unravel_index(10) == (SMCGridKind.cell, 10)


def test_grid_kinds(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert ds.ems.grid_kinds == [SMCGridKind.cell]


def test_default_grid_kind(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert ds.ems.default_grid_kind is SMCGridKind.cell


def test_get_grid_kind_and_size(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    assert ds.ems.get_grid_kind_and_size(ds['foo']) \
        == (SMCGridKind.cell, 20103)


def test_get_grid_kind_and_size_no_match(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    with pytest.raises(ValueError):
        ds.ems.get_grid_kind_and_size(ds['time'])


def test_make_linear(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    xarray.testing.assert_equal(ds.ems.make_linear(ds['foo']), ds['foo'])


def test_make_linear_no_match(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    with pytest.raises(ValueError):
        ds.ems.make_linear(ds['time'])


def test_get_selector_for_index(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    selector = ds.ems.selector_for_index((SMCGridKind, 20))
    assert selector == {ds.ems.topology.cell_dimension: 20}


def test_drop_geometry(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    topology = ds.ems.topology

    dropped_ds = ds.ems.drop_geometry()
    assert topology.cell_dimension in dropped_ds.dims
    assert topology.longitude_name not in dropped_ds.variables
    assert topology.latitude_name not in dropped_ds.variables
    assert topology.longitude_cell_size_factor_name not in dropped_ds.variables
    assert topology.latitude_cell_size_factor_name not in dropped_ds.variables


def test_polygons(
    datasets: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    polygons = ds.ems.polygons
    assert len(polygons) == ds.ems.topology.cell_count
    assert isinstance(polygons[0], Polygon)
    shapely.testing.assert_geometries_equal(
        polygons[0],
        Polygon([[100, -50], [102, -50], [102, -48], [100, -48]]))
    shapely.testing.assert_geometries_equal(
        polygons[-1],
        Polygon([[163, -7], [164, -7], [164, -6], [163, -6]]))


def test_clip(
    datasets: pathlib.Path,
    tmp_path: pathlib.Path,
):
    ds = xarray.open_dataset(datasets / 'smc.nc')
    tasmania = box(140, -44, 150, -39)
    clipped = ds.ems.clip(tasmania, work_dir=tmp_path)
    clipped.load()
    clipped.to_netcdf(tmp_path / 'clipped.nc')
    assert isinstance(clipped.ems, SMC)
    assert clipped.ems.topology.cell_count < ds.ems.topology.cell_count
    # TODO Test this geometry makes sense
